import std/[nre, options, os, sha1, strutils, times, unicode]
import console_progress, uuids
import ./db, ./libvlc, ./models

type
  WdiwtltLibrary* = ref object
    rootPath: string
    #autoDeletePeriodMs*: int # 1000 * 60 * 60 * 24 * 6 # one week
    db: WdiwtltDb
    vlc: LibVlcInstance

let FILENAME_PAT = re"(\d+[:-_ ]+)?(.+)$"
let RECOGNIZED_MEDIA_EXTENSIONS = [
  "3gp", "aac", "aif", "avi", "div", "flac", "flv", "h264", "m4a", "mid",
  "midi", "mka", "mkv", "mov", "mp3", "mp4a", "mpeg", "mpg", "mpg3", "mpg4",
  "mpga", "ogg", "vorb", "wav", "wma", "wma1", "wma2", "wmv", "wmv1", "wmv2",
  "x264"
]

iterator walkMediaFiles*(l: WdiwtltLibrary): string =
  for f in l.rootPath.walkDirRec(relative = true):
    let (_, name, ext) = f.splitFile
    if name.startsWith('.'): continue
    if not RECOGNIZED_MEDIA_EXTENSIONS.contains(ext.toLower): continue
    yield f

proc initLibrary*(rootPath: string, dbPath: string): WdiwtltLibrary =
  WdiwtltLibrary(rootPath: rootPath, db: loadDb(dbPath))

proc clean*(l: WdiwtltLibrary) =
  let staleDt = now() - 1.weeks

  l.db.removeEmptyAlbums
  l.db.removeEmptyArtists
  l.db.removeEmptyPlaylists
  l.db.pruneStalePlaylists(staleDt)
  l.db.pruneStaleBookmarks(staleDt)
  l.db.persist

proc initMediaFile*(
    l: WdiwtltLibrary,
    path: string,
    id = none[UUID](),
    hash = none[string]()
  ): tuple[
    mf: MediaFile,
    artistName: Option[string],
    albumName: Option[string],
    trackTotal: Option[int]
  ] =

  if not fileExists(path):
    raise newException(IOError, "file does not exist: '" & path & "'")

  let m = path.match(FILENAME_PAT)

  result = (
    MediaFile(
      comment: none[string](),
      dateAdded: now(),
      discNumber: none[string](),
      fileHash:
        if hash.isSome: hash.get
        else: $secureHashFile(path),
      filePath:
        if m.isSome: m.get.captures[2]
        else: path,
      id:
        if id.isSome: id.get
        else: genUUID(),
      imageUri: none[string](),
      lastPlayed: none[DateTime](),
      metaInfoSource: msFileLocation,
      name: path.splitFile.name,
      playCount: 0,
      presentLocally: true,
      trackNumber:
        if m.isSome and m.get.captures.contains(1):
          some(parseInt(m.get.captures[1]))
        else: none[int]()),
    none[string](),
    none[string](),
    none[int]())

  var media: VlcMedia
  try:
    media = l.vlc.mediaFromPath(path)
    media.parse

    let mName = media.getMeta(vmTitle)
    if not mName.isNil: result.mf.name = $mName

    let mDiscNo = media.getMeta(vmDiscNumber)
    if not mDiscNo.isNil: result.mf.discNumber = some($mDiscNo)

    let mImgUri = media.getMeta(vmArtworkURL)
    if not mImgUri.isNil: result.mf.imageUri = some($mImgUri)

    let mTrackNo = media.getMeta(vmTrackNumber)
    if not mTrackNo.isNil: result.mf.trackNumber = some(parseInt($mTrackNo))

    result.mf.metaInfoSource = msTagInfo

    let mAlbum = media.getMeta(vmAlbum)
    if not mAlbum.isNil: result.albumName = some($mAlbum)

    let mArtist = media.getMeta(vmArtist)
    if not mArtist.isNil: result.artistName = some($mArtist)

    let mTrackTotal = media.getMeta(vmTrackTotal)
    if not mTrackTotal.isNil:
      try: result.trackTotal = some(parseInt($mTrackTotal))
      except: result.trackTotal = none[int]()

  except: discard
  finally:
    if not media.isNil: media.release

proc scan*(l: WdiwtltLibrary) =

  var fileCount = 0
  for f in l.walkMediaFiles: fileCount += 1

  let progress = newProgress(stdout, fileCount)

  var curCount = 0
  for f in l.walkMediaFiles:
    progress.updateProgress(curCount, f[max(f.high - 15, 0)..f.high])

    # Skip this file if we already have a record of it
    let hash = $secureHashFile(f)
    var existingMf = l.db.findMediaFileByHash(hash)
    if existingMf.isSome: continue

    # Process this new file
    let (mf, artistsFromMeta, albumsFromMeta, trackTotal) = l.initMediaFile(f)
    l.db.add(mf)

    var allArtists = newSeq[Artist]()
    var allAlbums = newSeq[Album]()

    # Associate the file with its albums, creating album records if necessary
    if albumsFromMeta.isSome:
      let albumNames = albumsFromMeta.get.split(";")
      for name in albumNames:
        let existing = l.db.findAlbumsByName(name)
        allAlbums &= existing
        if existing.len > 0: l.db.add(existing[0], mf)
        else:
          let newAlbum = Album(
            id: genUUID(),
            name: name,
            imageUri: mf.imageUri,
            trackTotal:
              if trackTotal.isSome: trackTotal.get
              else: 0)
          allAlbums.add(newAlbum)
          l.db.add(newAlbum)
          l.db.add(newAlbum, mf)

    # Associate the file with its artists, creating artist records if necessary
    if artistsFromMeta.isSome:
      let artistNames = artistsFromMeta.get.split(";")
      for name in artistNames:
        let existing = l.db.findArtistsByName(name)
        allArtists &= existing
        if existing.len > 0: l.db.add(existing[0], mf)
        else:
          let newArtist = Artist(
            id: genUUID(),
            name: name,
            imageUri: mf.imageUri)
          allArtists.add(newArtist)
          l.db.add(newArtist)
          l.db.add(newArtist, mf)

    # Make sure we have association records between the artists and the albums
    if allAlbums.len > 0 and allArtists.len > 0:
      for album in allAlbums:
        for artist in allArtists:
          l.db.associate(allArtists[0], allAlbums[0])

  l.db.persist
